#include "AINetworkArc.h"
#include "AINetworkArcSelection.h"
#include "AINetworkNode.h"

#include <Map/MapCoordinateSystem.h>
#include <Map/MapEditor.h>
#include <Map/MapProjection.h>
#include <Math/CollisionUtils.h>

IMPLEMENT_CLASS(AINetworkArc,MapObject)

double AINetworkArc::Arc_Width=1.5;

AINetworkArc::AINetworkArc(TransactionManager* transactionManager,
                           AINetworkVertex::SmartRef source,
                           AINetworkVertex::SmartRef target,
                           bool isBidirectional,
                           bool isPushBackRoute,
                           wxString name):
        MapObject(transactionManager)
{
        TRANSACTIONOBJECT_STATEVAR(source)=source;
        TRANSACTIONOBJECT_STATEVAR(target)=target;
        TRANSACTIONOBJECT_STATEVAR(isBidirectional)=isBidirectional;
        TRANSACTIONOBJECT_STATEVAR(isPushBackRoute)=isPushBackRoute;
        TRANSACTIONOBJECT_STATEVAR(name)=name;
        source->AttachArc(this);
        target->AttachArc(this);
}

AINetworkArc::~AINetworkArc() {
}

void AINetworkArc::Paint(wxDC& dc, MapEditor* editor, const MapPass& pass) {
        if (pass!=AINetworkVertex::AINetworkPass) {
                return;
        }
        if (!IsPushBackRoute()) {
        	PaintArc(dc,editor,wxColour(192,192,0));
        } else {
		PaintArc(dc,editor,wxColour(192,0, 192));
        }
}

void AINetworkArc::PaintArc(wxDC& dc, MapEditor* editor, wxColour colour) {
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();
        MapProjection* proj=editor->GetProjection();
        
        Point3D sourcePos,targetPos;
        sourcePos=GetSource()->GetPosition();
        targetPos=GetTarget()->GetPosition();
        
        Point3D localSourcePos,localTargetPos;
        localSourcePos=proj->ConvertToLocal(sourcePos);
        localTargetPos=proj->ConvertToLocal(targetPos);
        
        wxPoint deviceSourcePos,deviceTargetPos;
        deviceSourcePos=cs.LogicalToDevice(localSourcePos);
        deviceTargetPos=cs.LogicalToDevice(localTargetPos);

        wxCoord width=cs.BoundedDeviceSizeX(3,10,Arc_Width);

        wxPen pen=wxPen(colour,width,wxSOLID);
        wxBrush brush=wxBrush(colour,wxSOLID);
        
        dc.SetPen(pen);
        dc.SetBrush(brush);
        dc.DrawLine(deviceSourcePos,deviceTargetPos);
                                                    
        if (!IsBidirectional()) {
                Point3D pts[3];
                
                GetArrowPoints(editor,pts);

                wxPoint devpts[3];
                
                for (int i=0;i<3;i++) {
                        devpts[i]=cs.LogicalToDevice(pts[i]);
                }
                dc.DrawPolygon(3,devpts);
        }
        
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);
}

void AINetworkArc::GetArrowPoints(MapEditor* editor, Point3D pts[3]) {
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();
        MapProjection* proj=editor->GetProjection();
        
        Point3D sourcePos,targetPos;
        sourcePos=GetSource()->GetPosition();
        targetPos=GetTarget()->GetPosition();
        
        Point3D localSourcePos,localTargetPos;
        localSourcePos=proj->ConvertToLocal(sourcePos);
        localTargetPos=proj->ConvertToLocal(targetPos);
        
        Point3D tip=(localSourcePos+localTargetPos*2.0)/3.0;
        double len=localSourcePos.distance3D(localTargetPos);
        double width=cs.BoundedLogicalSizeX(3,10,Arc_Width);
        Point3D v=(localTargetPos-localSourcePos)*width/len;
        Point3D w=Point3D(v.y(),-v.x(),0.0);
        
        pts[0]=tip;
        pts[1]=tip-v*4+w*2;
        pts[2]=tip-v*4-w*2;
}

bool AINetworkArc::Hit(MapEditor* editor, TDRectangle r) {
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();
        MapProjection* proj=editor->GetProjection();
        TDRectangle localBounds=GetLocalBounds(editor);
        if (!localBounds.Intersects(r))
                return false;
        
        Point3D sourcePos,targetPos;
        sourcePos=GetSource()->GetPosition();
        targetPos=GetTarget()->GetPosition();
        
        Point3D localSourcePos,localTargetPos;
        localSourcePos=proj->ConvertToLocal(sourcePos);
        localTargetPos=proj->ConvertToLocal(targetPos);
        
        double width=cs.BoundedLogicalSizeX(3,10,Arc_Width);
        
        double len=localSourcePos.distance3D(localTargetPos);
        Point3D v=localTargetPos-localSourcePos;
        Point3D w=Point3D(v.y(),-v.x(),0.0)*width/len/2.0;
        
        Point3D pts[4];
        
        pts[0]=localTargetPos+w;
        pts[1]=localSourcePos+w;
        pts[2]=localSourcePos-w;
        pts[3]=localTargetPos-w;
        
        return CollisionUtils::ConvexPolygonRectangleCollide(r, 4, pts);
}

MapSelection* AINetworkArc::CreateSelection(MapSelectionManager* manager) {
        return new AINetworkArcSelection(manager,this);
}

wxRect AINetworkArc::GetDeviceBounds(MapEditor* editor) {
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();
        MapProjection* proj=editor->GetProjection();
        
        Point3D sourcePos,targetPos;
        sourcePos=GetSource()->GetPosition();
        targetPos=GetTarget()->GetPosition();
        
        Point3D localSourcePos,localTargetPos;
        localSourcePos=proj->ConvertToLocal(sourcePos);
        localTargetPos=proj->ConvertToLocal(targetPos);
        
        wxPoint deviceSourcePos,deviceTargetPos;
        deviceSourcePos=cs.LogicalToDevice(localSourcePos);
        deviceTargetPos=cs.LogicalToDevice(localTargetPos);
        
        wxRect r=wxRect(deviceSourcePos,deviceTargetPos);
        
        wxCoord width=cs.BoundedDeviceSizeX(3,10,Arc_Width);
        r.Inflate(width/2+1);
        
        if (!IsBidirectional()) {
                Point3D pts[3];
                
                GetArrowPoints(editor,pts);
                
                for (int i=0;i<3;i++) {
                        wxPoint devpt;
                        devpt=cs.LogicalToDevice(pts[i]);
                        wxRect dr=wxRect(devpt,devpt);
                        r=r.Union(dr);
                }
        }
        return r;
}

TDRectangle AINetworkArc::GetGeographicBounds() {
        Point3D sourcePos,targetPos;
        sourcePos=GetSource()->GetPosition();
        targetPos=GetTarget()->GetPosition();
        
        return TDRectangle(sourcePos,targetPos);
}

void AINetworkArc::OnDetach() {
        Damage();
        GetSource()->DetachArc(this);
        GetTarget()->DetachArc(this);
}

AINetworkArc::SmartRef AINetworkArc::SplitAtCenter(MapEditor* editor) {
        MapProjection* proj=editor->GetProjection();
        
        Point3D sourcePos,targetPos;
        sourcePos=GetSource()->GetPosition();
        targetPos=GetTarget()->GetPosition();
        
        Point3D localSourcePos,localTargetPos;
        localSourcePos=proj->ConvertToLocal(sourcePos);
        localTargetPos=proj->ConvertToLocal(targetPos);
        
        Point3D localCenterPos=(localSourcePos+localTargetPos)/2.0;
        Point3D centerPos=proj->ConvertFromLocal(localCenterPos);
        
        AINetworkVertex::SmartRef centerVertex=new AINetworkNode(GetTransactionManager(),centerPos);
        AINetworkArc::SmartRef newarc=new AINetworkArc(GetTransactionManager(),centerVertex,GetTarget(),IsBidirectional(),IsPushBackRoute(),GetName());
        GetContainer()->AttachObject(centerVertex);
        GetContainer()->AttachObject(newarc);
        
        GetTarget()->DetachArc(this);
        SetTarget(centerVertex);
        centerVertex->AttachArc(this);
        
        return newarc;
}

