#include "AINetworkVertex.h"

#include "AINetworkArc.h"
#include <wx/listimpl.cpp>

#include <map>

typedef std::map<AINetworkVertex*,AINetworkArc*> ArcVertexMap;

WX_DEFINE_LIST(AINetworkVertexArcs);

IMPLEMENT_CLASS(AINetworkVertex,MapObject)

const MapPass AINetworkVertex::AINetworkPass(wxT("AI Network"));

AINetworkVertex::AINetworkVertex(TransactionManager* transactionManager,
                             Point3D position):
        MapObject(transactionManager)
{
        TRANSACTIONOBJECT_STATEVAR(position)=position;
}

AINetworkVertex::~AINetworkVertex() {
}

void AINetworkVertex::AttachArc(AINetworkArc* arc) {
        wxASSERT(TRANSACTIONOBJECT_STATEVAR(arcs).Find(arc)==NULL);
        BeforeWrite();
        arc->TransactionObject::Ref();
        TRANSACTIONOBJECT_STATEVAR(arcs).Append(arc);
        AfterWrite();
}

void AINetworkVertex::DetachArc(AINetworkArc* arc) {
        AINetworkVertexArcs::Node *arcNode;
        
        arcNode=TRANSACTIONOBJECT_STATEVAR(arcs).Find(arc);
        wxASSERT(arcNode!=NULL);
        BeforeWrite();
        TRANSACTIONOBJECT_STATEVAR(arcs).Erase(arcNode);
        arc->TransactionObject::UnRef();
        AfterWrite();
}

void AINetworkVertex::MergeArcsFrom(AINetworkVertex::SmartRef v) {
        ArcVertexMap mySourceMap,myTargetMap;
        
        AINetworkVertexArcs::Node* arcNode;
        AINetworkVertexArcs* arcs;
        
        arcs=&TRANSACTIONOBJECT_STATEVAR(arcs);
        for (arcNode=arcs->GetFirst();arcNode!=NULL;arcNode=arcNode->GetNext()) {
                AINetworkArc::SmartRef arc=arcNode->GetData();
                if (arc->GetSource()==this) {
                        myTargetMap[arc->GetTarget()]=arc;
                } else {
                        mySourceMap[arc->GetSource()]=arc;
                }
        }
        
        arcs=&v->TRANSACTIONOBJECT_STATEVAR(arcs);
        BeforeWrite();
        v->BeforeWrite();
        while (arcNode=arcs->GetFirst()) {
                AINetworkArc::SmartRef arc=arcNode->GetData();
                wxASSERT(arc->GetSource()==v || arc->GetTarget()==v);
                if (arc->GetSource()==v) {
                        // Don't introduce self-referential arcs
                        if (arc->GetTarget()==this) {
                                arc->RemoveFromContainer();
                                continue;
                        }
                        // Skip arcs to targets for which we also have an arc
                        if (myTargetMap.find(arc->GetTarget())!=myTargetMap.end()) {
                                arc->RemoveFromContainer();
                                continue;
                        }
                        // If we have an arc from the target of this arc,
                        // make our arc bidirectional
                        if (mySourceMap.find(arc->GetTarget())!=mySourceMap.end()) {
                                mySourceMap[arc->GetTarget()]->SetIsBidirectional(true);
                                arc->RemoveFromContainer();
                                continue;
                        }
                        
                        // This is definitely a new arc
                        arc->SetSource(this);
                } else if (arc->GetTarget()==v) {
                        // Don't introduce self-referential arcs
                        if (arc->GetSource()==this) {
                                arc->RemoveFromContainer();
                                continue;
                        }
                        // Skip arcs from sources for which we also have an arc
                        if (mySourceMap.find(arc->GetSource())!=mySourceMap.end()) {
                                arc->RemoveFromContainer();
                                continue;
                        }
                        // If we have an arc to the source of this arc,
                        // make our arc bidirectional
                        if (myTargetMap.find(arc->GetSource())!=myTargetMap.end()) {
                                myTargetMap[arc->GetSource()]->SetIsBidirectional(true);
                                arc->RemoveFromContainer();
                                continue;
                        }
                        
                        // This is definitely a new arc
                        arc->SetTarget(this);
                }
                arcs->DeleteNode(arcNode);
                TRANSACTIONOBJECT_STATEVAR(arcs).Append(arc);
        }
        v->AfterWrite();
        AfterWrite();
}

void AINetworkVertex::OnDetach() {
        // ensure that we won't trip when the arcs deregister themselves
        AINetworkVertexArcs arcs=TRANSACTIONOBJECT_STATEVAR(arcs);
        AINetworkVertexArcs::Node* arcNode;
        
        /* Detach all arcs connected to us */
        for (arcNode=arcs.GetFirst();arcNode!=NULL;arcNode=arcNode->GetNext()) {
                AINetworkArc::SmartRef arc=arcNode->GetData();
                arc->GetContainer()->DetachObject(arc);
        }
}

void AINetworkVertex::BeforeWrite() {
        MapObject::BeforeWrite();
        
        DamageArcs();
}

void AINetworkVertex::AfterWrite() {
        DamageArcs();
        
        MapObject::AfterWrite();
}

void AINetworkVertex::DamageArcs() {
        AINetworkVertexArcs& arcs=TRANSACTIONOBJECT_STATEVAR(arcs);
        AINetworkVertexArcs::Node* arcNode;
        
        for (arcNode=arcs.GetFirst();arcNode!=NULL;arcNode=arcNode->GetNext()) {
                AINetworkArc::SmartRef arc=arcNode->GetData();
                arc->Damage();
        }
}

AINetworkVertex::Memento::Memento(const AINetworkVertex& vertex):
        MapObject::Memento(vertex) {
        _savedState=vertex.TRANSACTIONOBJECT_STATE;
        AINetworkVertexArcs::Node *node;
        for (node=_savedState.arcs.GetFirst();node;node=node->GetNext()) {
                node->GetData()->TransactionObject::Ref();
        }
}

AINetworkVertex::Memento::~Memento() {
        AINetworkVertexArcs::Node *node;
        for (node=_savedState.arcs.GetFirst();node;node=node->GetNext()) {
                node->GetData()->TransactionObject::UnRef();
        }
}

void AINetworkVertex::RestoreMemento(wxObject* object) {
        MapObject::RestoreMemento(object);
        Memento* memento=dynamic_cast<Memento*>(object);
        wxASSERT(memento);
        
        AINetworkVertexArcs::Node* arcNode;
        
        for (arcNode=memento->_savedState.arcs.GetFirst();arcNode!=NULL;arcNode=arcNode->GetNext()) {
                arcNode->GetData()->TransactionObject::Ref();
        }
        for (arcNode=TRANSACTIONOBJECT_STATEVAR(arcs).GetFirst();arcNode!=NULL;arcNode=arcNode->GetNext()) {
                arcNode->GetData()->TransactionObject::UnRef();
        }
        TRANSACTIONOBJECT_STATE=memento->_savedState;
        Damage();
}

wxAINetworkVertexArcsNode *AINetworkVertex::getFirstArc() { 
        AINetworkVertexArcs& arcs=TRANSACTIONOBJECT_STATEVAR(arcs);
        wxAINetworkVertexArcsNode* arcNode;
        
        arcNode=arcs.GetFirst();
        return arcNode;
};

