#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "PavementMapObject.h"
#include "PavementSelection.h"
#include <Map/MapEditor.h>
#include <Map/MapCoordinateSystem.h>
#include <Map/MapProjection.h>
#include <Math/sg_geodesy.h>
#include <Math/CollisionUtils.h>

IMPLEMENT_CLASS(PavementMapObject,MapObject)

const MapPass PavementMapObject::LightingPass(wxT("Lighting"));

PavementMapObject::PavementMapObject(Airport::SmartRef airport,
                                     Point3D center,
                                     double length,
                                     double width,
                                     double heading,
                                     Surface surface,
                                     float smoothness):
        MapObject(airport->GetTransactionManager()),
        _airport(airport)
{
        TRANSACTIONOBJECT_STATEVAR(_center)=center;
        TRANSACTIONOBJECT_STATEVAR(_length)=length;
        TRANSACTIONOBJECT_STATEVAR(_width)=width;
        TRANSACTIONOBJECT_STATEVAR(_heading)=heading;
        TRANSACTIONOBJECT_STATEVAR(_surface)=surface;
        TRANSACTIONOBJECT_STATEVAR(_smoothness)=smoothness;
        _airport->RegisterChangeListener(this);
        UpdateCorners();
}

PavementMapObject::~PavementMapObject() {
        _airport->UnregisterChangeListener(this);
}

/* Paint this object using the given device context into the given editor */
void PavementMapObject::Paint(wxDC& dc, MapEditor* editor, const MapPass& pass) {
        if (pass!=StandardPass)
                return;
        wxPoint corners[4];
        MapProjection* proj=editor->GetProjection();
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        for (int i=0;i<4;i++)
                corners[i]=cs->LogicalToDevice(proj->ConvertToLocal(_endCorners[i]));
        
        dc.SetPen(*wxBLACK_PEN);
        dc.SetBrush(wxBrush(GetSurfaceColour()));
        dc.DrawPolygon(4,corners);
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);
}

/* Return TRUE iff this object intersects the given hit-TDRectangle r to list.
 * r is given in the local coordinate system of editor. */
bool PavementMapObject::Hit(MapEditor* editor, TDRectangle r) {
        TDRectangle bounds=GetLocalBounds(editor);
        if (!bounds.Intersects(r))
                return false;
        MapProjection* proj=editor->GetProjection();
        Point3D corners[4];
        for (unsigned int i=0;i<4;i++)
                corners[i]=proj->ConvertToLocal(_endCorners[i]);
        
        return CollisionUtils::ConvexPolygonRectangleCollide(r, 4, corners);
}

wxRect PavementMapObject::GetDeviceBounds(MapEditor* editor) {
        wxCoord top,bottom,left,right;
        
        MapProjection* proj=editor->GetProjection();
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        
        wxPoint corner;
        corner=cs->LogicalToDevice(proj->ConvertToLocal(_endCorners[0]));
        left=right=corner.x;
        top=bottom=corner.y;
        
        for (int i=1;i<4;i++) {
                corner=cs->LogicalToDevice(proj->ConvertToLocal(_endCorners[i]));
                left=std::min(left,corner.x);
                right=std::max(right,corner.x);
                top=std::min(top,corner.y);
                bottom=std::max(bottom,corner.y);
        }
        return wxRect(left,top,right-left,bottom-top);
}

/* Return the geographic bounds (lat/lon) */
TDRectangle PavementMapObject::GetGeographicBounds() {
        TDRectangle rect;
        for (int i=0;i<4;i++)
                rect=TDRectangle::Union(rect,TDRectangle(_endCorners[i]));
        return rect;
}

wxColour PavementMapObject::GetSurfaceColour() const {
        switch (GetSurface()) {
        case Surface_Asphalt:
        case Surface_Helipad_Asphalt:
        case Surface_Twy_Asphalt_yellow_hold:
                return wxColour(128,128,128);
        case Surface_Concrete:
        case Surface_Helipad_Concrete:
        case Surface_Twy_Concrete_yellow_hold:
                return *wxLIGHT_GREY;
        case Surface_Turf:
        case Surface_Helipad_Turf:
                return *wxGREEN;
        case Surface_Dirt:
        case Surface_Helipad_Dirt:
                return wxColour(100,80,40);
        case Surface_Gravel:
                return wxColour(64,64,0);
        case Surface_Lakebed:
                return wxColour(230,230,100);
        case Surface_Water:
                return *wxBLUE;
        }
}

void PavementMapObject::UpdateCorners() {
        double halflength[2];
        
        halflength[0]=halflength[1]=GetLength()*0.3048/2.0;
        
        CalculateRectCorners(GetCenter().lat(),GetCenter().lon(),
                             GetAirport()->GetElevation()*0.3048,
                             GetHeading(),
                             halflength,
                             GetWidth()*0.3048,
                             _endCenters,
                             _endCorners);
        
}

void PavementMapObject::CalculateRectCorners(double centerlat, double centerlon,
                                             double elev,
                                             double heading,
                                             double halflength[2],
                                             double width,
                                             Point3D centerEnds[2],
                                             Point3D corners[4])
{
        double az;
        double endlat,endlon;
        double cornerlat,cornerlon;
        double angle,angle2;
        int i,j;
        
        // End 0 is the designated end, End 1 is the opposite end
        // corner 0 is left corner on designated end, corners are ordered clockwise
        for (angle=heading+180.0,i=0;i<2;i++,angle+=180.0) {
                // Calculate the center vertex
                geo_direct_wgs_84(elev,centerlat,centerlon,
                                  angle,
                                  halflength[i],
                                  &endlat,&endlon,&az);
                centerEnds[i]=Point3D(endlon,endlat,elev);
                
                // Calculate corner points
                for (angle2=-90.0,j=0;j<2;j++,angle+=180.0) {
                        geo_direct_wgs_84(elev,endlat,endlon,
                                          angle+angle2,
                                          width/2.0,
                                          &cornerlat,&cornerlon,&az);
                        corners[2*i+j]=Point3D(cornerlon,cornerlat,elev);
                }
        }
}
