#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "PavementSelection.h"
#include "PavementMapObject.h"
#include <Map/MapEditor.h>
#include <Map/MapSelectionManager.h>
#include <Map/MapCoordinateSystem.h>
#include <Map/MapProjection.h>
#include <math.h>
#include "PavementSelectionEndHandle.h"
#include "PavementSelectionSideHandle.h"

IMPLEMENT_CLASS(PavementSelection,MapSelection);

PavementSelection::PavementSelection(MapSelectionManager* manager, PavementMapObject* content):
        MapSelection(manager,content) {
}

PavementSelection::~PavementSelection() {
}

MapSelectionHandle* PavementSelection::HitHandle(Point3D p) {
        if (!HandlesActive())
                return (MapSelectionHandle*)NULL;
        MapEditor* editor=GetSelectionManager()->GetEditor();
        MapProjection* proj=editor->GetProjection();
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        
        Point3D rectsize=GetHandleBoxLogicalSize();
        for (unsigned int i=0;i<2;i++) {
                Point3D endpoint=proj->ConvertToLocal(GetContent()->GetEndCenters()[i]);
                Point3D delta=p-endpoint;
                if (endpoint.distance3Dsquared(p)*4.0<(rectsize.x()*rectsize.x())) {
                        return new PavementSelectionEndHandle(this,i);
                }
        }
        
        double dl=5,length,width;
        int side;
        
        length=GetContent()->GetLength();
        width=GetContent()->GetWidth();
        if (dl>length/2)
                dl=length*0.9/2;
        if (dl>width/2)
                dl=width*0.9/2;
        side=GetHitSide(p,GetContent()->GetEndCorners(),dl);
        if (side>=0) {
                return new PavementSelectionSideHandle(this,side);
        }
        
        return MapSelection::HitHandle(p);
}

int PavementSelection::GetHitSide(const Point3D localPoint, const Point3D corners[4], double dl) {
        MapEditor* editor=GetSelectionManager()->GetEditor();
        MapProjection* proj=editor->GetProjection();

        Point3D localCorners[4];
        for (unsigned int i=0;i<4;i++)
                localCorners[i]=proj->ConvertToLocal(corners[i]);

        double a,b,cx,cy,d,s;

        for (unsigned int i=0;i<4;i++) {
                Point3D *cornerA,*cornerB;
                cornerA=localCorners+i;
                if (i<3)
                        
                        cornerB=cornerA+1;
                else
                        cornerB=localCorners;
                a=cornerA->y()-cornerB->y();
                b=cornerB->x()-cornerA->x();
                cx=(cornerA->x()+cornerB->x())/2;
                cy=(cornerA->y()+cornerB->y())/2;
                d=a*(localPoint.x()-cornerA->x())+b*(localPoint.y()-cornerA->y());
                s=-b*(localPoint.x()-cx)+a*(localPoint.y()-cy);
                d=d*d/(a*a+b*b);
                s=s/(a*a+b*b);
                if (fabs(s)>=0.5)
                        continue;
                if (d<dl)
                        return i;
        }
        return -1;
}

wxRect PavementSelection::GetDeviceBounds() const {
        wxRect r=MapSelection::GetDeviceBounds();
        if (HandlesActive()) {
                // make sure the side handles are contained (pen width 3)
                r.Inflate(3,3);
                
                MapEditor* editor=GetSelectionManager()->GetEditor();
                MapProjection* proj=editor->GetProjection();
                MapCoordinateSystem& cs=editor->GetCoordinateSystem();
                
                wxSize rectsize=GetHandleBoxDeviceSize();
                for (int i=0;i<2;i++) {
                        wxPoint centerend;
                        centerend=cs.LogicalToDevice(proj->ConvertToLocal(GetContent()->GetEndCenters()[i]));
                        // make sure the actual line is contained (pen width 3)
                        wxRect handleRect=wxRect(wxPoint(centerend.x-rectsize.x/2-1,
                                                         centerend.y-rectsize.y/2-1),
                                                 wxSize(rectsize.x+3,rectsize.y+3));
                        r=r.Union(handleRect);
                }
        }
        wxSize handleSize=GetHandleBoxDeviceSize();
        r.Inflate(handleSize.x,handleSize.y);
        return r;
}

void PavementSelection::DeltaDrag(Point3D delta) {
        MapProjection* projection=GetSelectionManager()->GetEditor()->GetProjection();
        Point3D localCenter=projection->ConvertToLocal(GetContent()->GetCenter());
        localCenter+=delta;
        Point3D newCenter=projection->ConvertFromLocal(localCenter);
        GetContent()->SetCenter(newCenter);
}

void PavementSelection::Paint(wxDC& dc) {
        if (HandlesActive()) {
                PaintHandles(dc);
        }
        MapEditor* editor=GetSelectionManager()->GetEditor();
        MapProjection* proj=editor->GetProjection();
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();

        wxSize rectsize=GetHandleBoxDeviceSize();
        const Point3D *corners=GetContent()->GetEndCorners();
        wxPoint devCorner;
        
        wxPen markingPen=wxPen(Selection_Marking_Colour);
        wxBrush markingBrush=wxBrush(Selection_Marking_Colour);
        dc.SetPen(markingPen);
        dc.SetBrush(markingBrush);
        for (int i=0;i<4;i++) {
                devCorner=cs.LogicalToDevice(proj->ConvertToLocal(corners[i]));
                dc.DrawRectangle(devCorner.x-rectsize.x/2,
                                 devCorner.y-rectsize.y/2,
                                 rectsize.x,rectsize.y);
        }
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);
}

void PavementSelection::ProvidePropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass) {
        MapSelection::ProvidePropertiesForClass(propgrid,targetClass);
        if (!targetClass->IsKindOf(CLASSINFO(PavementSelection)))
                return;
        
        wxPGChoices surfaceChoices;
        surfaceChoices.Add(wxT("Asphalt"),PavementMapObject::Surface_Asphalt);
        surfaceChoices.Add(wxT("Concrete"),PavementMapObject::Surface_Concrete);
        surfaceChoices.Add(wxT("Turf"),PavementMapObject::Surface_Turf);
        surfaceChoices.Add(wxT("Dirt"),PavementMapObject::Surface_Dirt);
        surfaceChoices.Add(wxT("Gravel"),PavementMapObject::Surface_Gravel);
        surfaceChoices.Add(wxT("Dry Lakebed"),PavementMapObject::Surface_Lakebed);
        surfaceChoices.Add(wxT("Water"),PavementMapObject::Surface_Water);

        propgrid->Append(wxEnumProperty(wxT("Surface"),wxT("PavementSelection_Surface"),surfaceChoices));
        propgrid->Append(wxFloatProperty(wxT("Smoothness"),wxT("PavementSelection_Smoothness")));
        propgrid->AppendCategory(wxT("Pavement Geometry"));
        propgrid->Append(wxFloatProperty(wxT("Center Latitude (deg)"),wxT("PavementSelection_Geometry_Center_Lat")));
        propgrid->Append(wxFloatProperty(wxT("Center Longitude (deg)"),wxT("PavementSelection_Geometry_Center_Lon")));
        propgrid->Append(wxFloatProperty(wxT("True Heading (deg)"),wxT("PavementSelection_Geometry_Heading")));
        propgrid->Append(wxFloatProperty(wxT("Length (ft)"),wxT("PavementSelection_Geometry_Length")));
        propgrid->Append(wxFloatProperty(wxT("Width (ft)"),wxT("PavementSelection_Geometry_Width")));
}

void PavementSelection::UpdatePropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass) {
        MapSelection::UpdatePropertiesForClass(propgrid,targetClass);
        if (!targetClass->IsKindOf(CLASSINFO(PavementSelection)))
                return;
        Point3D center=GetContent()->GetCenter();
        float heading=GetContent()->GetHeading();
        float length=GetContent()->GetLength();
        float width=GetContent()->GetWidth();
        propgrid->SetPropertyValue(wxT("PavementSelection_Surface"),(int)GetContent()->GetSurface());
        propgrid->SetPropertyValue(wxT("PavementSelection_Smoothness"),GetContent()->GetSmoothness());
        propgrid->SetPropertyValue(wxT("PavementSelection_Geometry_Center_Lat"),center.lat());
        propgrid->SetPropertyValue(wxT("PavementSelection_Geometry_Center_Lon"),center.lon());
        propgrid->SetPropertyValue(wxT("PavementSelection_Geometry_Heading"),heading);
        propgrid->SetPropertyValue(wxT("PavementSelection_Geometry_Length"),length);
        propgrid->SetPropertyValue(wxT("PavementSelection_Geometry_Width"),width);
}

void PavementSelection::MarkDifferingPropertiesForClass(wxPropertyGrid* propgrid, wxClassInfo* targetClass, MapSelection* firstSelection) {
        MapSelection::MarkDifferingPropertiesForClass(propgrid,targetClass,firstSelection);
        if (!targetClass->IsKindOf(CLASSINFO(PavementSelection)))
                return;
        PavementMapObject* myContent=GetContent();
        PavementMapObject* firstContent=wxDynamicCast(firstSelection,PavementSelection)->GetContent();
        if (myContent->GetSurface()!=firstContent->GetSurface())
                propgrid->SetPropertyUnspecified(wxT("PavementSelection_Surface"));
        if (myContent->GetSmoothness()!=firstContent->GetSmoothness())
                propgrid->SetPropertyUnspecified(wxT("PavementSelection_Smoothness"));
        if (myContent->GetCenter().lat()!=firstContent->GetCenter().lat())
                propgrid->SetPropertyUnspecified(wxT("PavementSelection_Geometry_Center_Lat"));
        if (myContent->GetCenter().lon()!=firstContent->GetCenter().lon())
                propgrid->SetPropertyUnspecified(wxT("PavementSelection_Geometry_Center_Lon"));
        if (myContent->GetHeading()!=firstContent->GetHeading())
                propgrid->SetPropertyUnspecified(wxT("PavementSelection_Geometry_Heading"));
        if (myContent->GetLength()!=firstContent->GetLength())
                propgrid->SetPropertyUnspecified(wxT("PavementSelection_Geometry_Length"));
        if (myContent->GetWidth()!=firstContent->GetWidth())
                propgrid->SetPropertyUnspecified(wxT("PavementSelection_Geometry_Width"));
}

void PavementSelection::UpdateObjectFromProperty(wxPGProperty* property) {
        const wxString name=property->GetName();
        wxVariant value=property->GetValueAsVariant();
        if (name.Cmp(wxT("PavementSelection_Surface"))==0) {
                GetContent()->SetSurface((PavementMapObject::Surface)value.GetLong());
        } else if (name.Cmp(wxT("PavementSelection_Smoothness"))==0) {
                GetContent()->SetSmoothness(value.GetDouble());
        } else if (name.Cmp(wxT("PavementSelection_Geometry_Center_Lat"))==0) {
                Point3D center=GetContent()->GetCenter();
                center.setlat(value.GetDouble());
                GetContent()->SetCenter(center);
        } else if (name.Cmp(wxT("PavementSelection_Geometry_Center_Lon"))==0) {
                Point3D center=GetContent()->GetCenter();
                center.setlon(value.GetDouble());
                GetContent()->SetCenter(center);
        } else if (name.Cmp(wxT("PavementSelection_Geometry_Heading"))==0) {
                GetContent()->SetHeading(value.GetDouble());
        } else if (name.Cmp(wxT("PavementSelection_Geometry_Length"))==0) {
                GetContent()->SetLength(value.GetDouble());
        } else if (name.Cmp(wxT("PavementSelection_Geometry_Width"))==0) {
                GetContent()->SetWidth(value.GetDouble());
        } else {
                MapSelection::UpdateObjectFromProperty(property);
        }
}

void PavementSelection::PaintHandles(wxDC& dc) {
        MapEditor* editor=GetSelectionManager()->GetEditor();
        MapProjection* proj=editor->GetProjection();
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        
        wxPoint corners[4];
        for (int i=0;i<4;i++)
                corners[i]=cs->LogicalToDevice(proj->ConvertToLocal(GetContent()->GetEndCorners()[i]));

        wxSize rectsize=GetHandleBoxDeviceSize();
        wxPoint centerends[2];
        for (int i=0;i<2;i++)
                centerends[i]=cs->LogicalToDevice(proj->ConvertToLocal(GetContent()->GetEndCenters()[i]));
        
        wxPen handlePen=wxPen(Selection_Handle_Colour);
        handlePen.SetWidth(3);
        dc.SetPen(handlePen);
        dc.SetBrush(*wxTRANSPARENT_BRUSH);

        /* draw edge handles */
        dc.DrawPolygon(4,corners);

        /* draw endpoint handles */
        for (int i=0;i<2;i++) {
                dc.DrawCircle(centerends[i],rectsize.x/2);
        }
        
        dc.SetBrush(wxNullBrush);
        dc.SetPen(wxNullPen);
}

