#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "PavementSelectionSideHandle.h"
#include "PavementSelection.h"

#include <Map/MapEditor.h>
#include <Map/MapSelectionManager.h>

#include <Math/sg_geodesy.h>

#include <Map/MapProjection.h>

IMPLEMENT_CLASS(PavementSelectionSideHandle,MapSelectionHandle);

PavementSelectionSideHandle::PavementSelectionSideHandle(MapSelection* selection, unsigned int side):
        MapSelectionHandle(selection),
        _side(side)
{
}

PavementSelectionSideHandle::~PavementSelectionSideHandle()
{
}

void PavementSelectionSideHandle::StartDrag(Point3D startPosition) {
        PavementSelection* selection=(PavementSelection*)GetSelection();
        PavementMapObject* object=selection->GetContent();
        
        StartDrag(startPosition,
                  object->GetEndCorners(),
                  _side);
}

void PavementSelectionSideHandle::StartDrag(Point3D startPosition, const Point3D* corners, int offset) {
        PavementSelection* selection=(PavementSelection*)GetSelection();
        MapEditor* editor=selection->GetSelectionManager()->GetEditor();
        MapProjection* projection=editor->GetProjection();
        
        Point3D localcorners[2];
        for (int i=0;i<2;i++)
                localcorners[i]=projection->ConvertToLocal(corners[(offset+i)%4]);
        Point3D localcenter;
        localcenter=(localcorners[0]+localcorners[1])/2;
        _handleDelta=startPosition-localcenter;
}

void PavementSelectionSideHandle::Drag(Point3D currentPosition) {
        PavementSelection* selection=(PavementSelection*)GetSelection();
        PavementMapObject* object=selection->GetContent();
        MapEditor* editor=selection->GetSelectionManager()->GetEditor();
        MapProjection* projection=editor->GetProjection();
        
        Point3D localcenters[2];
        double length;
        
        length=GetNewLength(currentPosition,
                            object->GetEndCorners(),
                            _side,
                            localcenters);
        
        
        /*
         * Determine new center
         *
         * Here we assume that the user wants only the dragged end
         * to move _in the current projection_. This solution
         * emulates manually moving the center by half the way after
         * changing the length. It also seems to work quite well.
         */
        // FIXME: Avoid the pavement moving around when it is collapsed to zero length
        Point3D localCenter=(localcenters[0]+localcenters[1])/2.0;
        Point3D center=projection->ConvertFromLocal(localCenter);
        
        object->SetCenter(center);
        
        if (_side==0 || _side==2)
                object->SetLength(length);
        else
                object->SetWidth(length);
}

double PavementSelectionSideHandle::GetNewLength(Point3D currentPosition,
                                                 const Point3D* corners,
                                                 int offset,
                                                 Point3D localcenters[2])
{
        PavementSelection* selection=(PavementSelection*)GetSelection();
        PavementMapObject* object=selection->GetContent();
        MapEditor* editor=selection->GetSelectionManager()->GetEditor();
        MapProjection* projection=editor->GetProjection();
        
        // determine the constraint vector
        Point3D localcorners[4];
        for (int i=0;i<4;i++)
                localcorners[i]=projection->ConvertToLocal(corners[(offset+i)%4]);
        localcenters[0]=(localcorners[0]+localcorners[1])/2;
        localcenters[1]=(localcorners[2]+localcorners[3])/2;
        
        Point3D cv=localcenters[0]-localcenters[1];
        double len=localcenters[1].distance3D(localcenters[0]);
        cv/=len;

        /*
         * Determine the constrained position
         */
        Point3D delta=currentPosition-_handleDelta-localcenters[1];
        double deltaL=delta.x()*cv.x()+delta.y()*cv.y();
        Point3D cstrDelta=deltaL*cv;
        
        localcenters[0]=localcenters[1]+cstrDelta;
        
        Point3D newcenters[2];
        for (int i=0;i<2;i++)
                newcenters[i]=projection->ConvertFromLocal(localcenters[i]);
        
        /*
         * Determine new "width"
         */
        double alt=object->GetAirport()->GetElevation()*0.3048;
        double az1, az2, length;
        geo_inverse_wgs_84( alt,
                            newcenters[1].lat(), newcenters[1].lon(),
                            newcenters[0].lat(), newcenters[0].lon(),
                            &az1, &az2, &length );
        length/=0.3048;
        
        /* Return a negative length if the vector flipped over */
        return (deltaL>=0.0?length:-length);
}

wxString PavementSelectionSideHandle::EndDrag() {
        // FIXME: should delete this object now
        if (_side==0 || _side==2)
                return wxT("Changing pavement length");
        else
                return wxT("Changing pavement width");
}
