#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "PointMapObject.h"
#include "PointObjectSelection.h"
#include <Map/MapEditor.h>
#include <Map/MapCoordinateSystem.h>
#include <Map/MapSelection.h>
#include <Map/MapProjection.h>
#include <Math/point3d.h>
#include <Math/CollisionUtils.h>

IMPLEMENT_CLASS(PointMapObject,MapObject)

PointMapObject::PointMapObject(TransactionManager* transactionManager,
                               Point3D position,
                               const wxPen outerRingPen,
                               const wxPen innerRingPen,
                               const wxBrush innerRingBrush):
        MapObject(transactionManager),
        _outerRingPen(outerRingPen),
        _innerRingPen(innerRingPen),
        _innerRingBrush(innerRingBrush) {
                TRANSACTIONOBJECT_STATEVAR(_position)=position;
}

PointMapObject::~PointMapObject() {
}

/* Paint this object using the given device context into the given editor */
void PointMapObject::Paint(wxDC& dc, MapEditor* editor, const MapPass& pass) {
        if (pass!=StandardPass)
                return;
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        Point3D localPos=editor->GetProjection()->ConvertToLocal(GetPosition());
        wxPoint center=cs->LogicalToDevice(localPos);
        
        /* Determine the size of inner and outer ring */
        wxSize innerSize,addOuterSize;
        innerSize=cs->LogicalToDeviceRel(Point3D(12,12,0));
        addOuterSize=cs->LogicalToDeviceRel(Point3D(8,8,0));
        
        /* Draw outer transparent ring */
        dc.SetPen(_outerRingPen);
        dc.SetBrush(*wxTRANSPARENT_BRUSH);
        dc.DrawCircle(center,innerSize.GetWidth()+addOuterSize.GetWidth());
        
        /* Draw inner filled ring */
        dc.SetPen(_innerRingPen);
        dc.SetBrush(_innerRingBrush);
        dc.DrawCircle(center,innerSize.GetWidth());
        
        dc.SetPen(wxNullPen);
        dc.SetBrush(wxNullBrush);
}

/* Return TRUE iff this object intersects the given hit-TDRectangle r to list.
 * r is given in the local coordinate system of editor. */
bool PointMapObject::Hit(MapEditor* editor, TDRectangle rect) {
        Point3D localPos=editor->GetProjection()->ConvertToLocal(GetPosition());
        return CollisionUtils::IntersectsCircle(rect,localPos,20.0);
}

wxRect PointMapObject::GetDeviceBounds(MapEditor* editor) {
        MapCoordinateSystem* cs=&editor->GetCoordinateSystem();
        Point3D localPos=editor->GetProjection()->ConvertToLocal(GetPosition());
        wxPoint center=cs->LogicalToDevice(localPos);

        /* Determine the size of inner and outer ring */
        wxSize innerSize,addOuterSize;
        innerSize=cs->LogicalToDeviceRel(Point3D(12,12,0));
        addOuterSize=cs->LogicalToDeviceRel(Point3D(8,8,0));
        wxCoord totalRadius;
        totalRadius=innerSize.GetWidth()+addOuterSize.GetWidth()+_outerRingPen.GetWidth();
        
        return wxRect(center.x-totalRadius,center.y-totalRadius,
                      2*totalRadius,2*totalRadius);
}

/* Return the geographic bounds (lat/lon) */
TDRectangle PointMapObject::GetGeographicBounds() {
        /* We only return the center, as we can't give the size of the
         * symbol without a projection
         */
        return TDRectangle(GetPosition());
}

