#ifndef MapEditorDecoration_h
#define MapEditorDecoration_h

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
// Include your minimal set of headers here, or wx.h
#include <wx/wx.h>
#endif

class MapEditor;

class MapEditorDecoration;

WX_DECLARE_LIST(MapEditorDecoration,MapEditorDecorationList);

/**
 * \brief A MapEditorDecoration represents a graphical decoration of a map.
 *
 * A decoration currenly differs from a layer in that its extent is typically
 * dependent on the configuration of the MapCoordinateSystem. In case of a grid,
 * e.g., the grid covers exactly the current display area.
 *
 * Therefore decorations must be explicitly informed about changes in the
 * coordinate system which do not automatically result in a complete redraw of
 * the whole display area, such as scrolling.
 *
 * A decoration is also only attached to a single editor, in contrast to a
 * MapLayer, which may be attached to multiple editors.
 *
 * Decorations can also be enabled or disabled. When a decoration is disabled,
 * it is not painted.
 *
 * \todo Try to join this with MapLayer to end the special handling of
 * decorations in MapEditor.
 */
class MapEditorDecoration: public wxObject {
        
        DECLARE_ABSTRACT_CLASS(MapEditorDecoration)
        
public:
	MapEditorDecoration(MapEditor* editor);
        
        /**
         * \brief Destroy this decoration.
         *
         * This may only be called from MapEditor.
         */
        virtual ~MapEditorDecoration();
        
        /**
         * \brief Paint the decoration.
         *
         * \param dc    The device context to paint on.
         */
        virtual void Paint(wxDC& dc)=0;
        
        /**
         * \brief React on scrolling the viewport.
         *
         * For example, simple scrolling typically damages a
         * grid as the areas containing the latitude and longitude
         * labels are not overwritten. The grid can therefore
         * demand repainting of these areas.
         */
        virtual void Scroll(wxSize) {
        }
        
        MapEditor* GetEditor() const { return _editor; }
        
        bool IsEnabled() const { return _enabled; }
        void Enable() { _enabled=true; }
        void Disable() { _enabled=false; }
protected:
        MapEditor* _editor;
        
        bool _enabled;
};

#endif

