#include "PavementInsertionMode.h"

#include <Airport/Airport.h>
#include <Airport/PavementMapObject.h>
#include <Airport/RunwayMapObject.h>
#include <Airport/TaxiwayMapObject.h>
#include <Map/MapCoordinateSystem.h>
#include <Map/MapEditor.h>
#include <Map/MapProjection.h>
#include <Map/MapSelectionManager.h>
#include <Math/sg_geodesy.h>

IMPLEMENT_CLASS(PavementInsertionMode,Mode)

BEGIN_EVENT_TABLE(PavementInsertionMode,Mode)
        EVT_LEFT_DOWN(PavementInsertionMode::OnLeftDown)
        EVT_LEFT_UP(PavementInsertionMode::OnLeftUp)
        EVT_MOTION(PavementInsertionMode::OnMouseMotion)
        EVT_PAINT(PavementInsertionMode::OnPaint)
END_EVENT_TABLE()

PavementInsertionMode::PavementInsertionMode(Airport::SmartRef airport,InsertionMode mode):
        _airport(airport),
        _mode(mode),
        _active(false),
        _pointCount(0)
{
}

PavementInsertionMode::~PavementInsertionMode() {
}

/* Leave this mode */
void PavementInsertionMode::Leave(bool force) {
        if (_pointCount>0) {
        MapEditor* editor=GetEditor();
                MapCoordinateSystem& cs=editor->GetCoordinateSystem();
                MapProjection* proj=editor->GetProjection();
                TransactionManager* tmgr=editor->GetTransactionManager();
                
                tmgr->Abort();
                
                // clear the centerline indicator
                wxRect devRect;
                wxPoint pts[2];
                Point3D localFirstPoint=proj->ConvertToLocal(_centerEnds[0]);
                Point3D localSecondPoint=proj->ConvertToLocal(_centerEnds[1]);
                pts[0]=cs.LogicalToDevice(localFirstPoint);
                pts[1]=cs.LogicalToDevice(localSecondPoint);
                
                devRect=wxRect(pts[0],pts[1]);
                GetEditor()->DamageRect(devRect);
        }
        Mode::Leave(force);
}

/* Get a string for the status bar. */
wxString PavementInsertionMode::GetStatus() {
        switch (_pointCount) {
        case 0:
                return wxT("Click to define the center of the first end.");
        case 1:
                return wxT("Click to define the center of the second end.");
        case 2:
                return wxT("Move the mouse to define the width and click to finish.");
        }
        return Mode::GetStatus();
}

void PavementInsertionMode::OnLeftDown(wxMouseEvent& evt) {
        if (evt.AltDown() || evt.ControlDown()) {
                evt.Skip();
                return;
        }
        _active=true;
}

void PavementInsertionMode::OnPaint(wxPaintEvent& evt) {
        if (_pointCount!=1) {
                evt.Skip();
                return;
        }
        wxClientDC dc(GetEditor()->GetDrawingWindow());
        MapEditor* editor=GetEditor();
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();
        MapProjection* proj=editor->GetProjection();
        
        wxPoint endpts[2];
        endpts[0]=cs.LogicalToDevice(proj->ConvertToLocal(_centerEnds[0]));
        endpts[1]=cs.LogicalToDevice(proj->ConvertToLocal(_centerEnds[1]));
        
        wxPen pen(*wxBLACK,1,wxDOT);
        dc.SetPen(pen);
        
        dc.DrawLine(endpts[0],endpts[1]);

        dc.SetPen(wxNullPen);
        
        // Pass on to other modes
        evt.Skip();
}

void PavementInsertionMode::OnMouseMotion(wxMouseEvent& evt) {
        MapEditor* editor=GetEditor();
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();
        MapProjection* proj=editor->GetProjection();
        wxPoint devPoint=evt.GetPosition();
        Point3D localPoint=cs.DeviceToLogical(devPoint);
        
        switch (_pointCount) {
                case 0:
                        break;
                case 1: {
                        /* update the drawn centerline */
                        wxRect devRect;
                        wxPoint pts[2];
                        Point3D localFirstPoint=proj->ConvertToLocal(_centerEnds[0]);
                        Point3D localSecondPoint=proj->ConvertToLocal(_centerEnds[1]);
                        pts[0]=cs.LogicalToDevice(localFirstPoint);
                        pts[1]=cs.LogicalToDevice(localSecondPoint);
                        
                        devRect=wxRect(pts[0],pts[1]);
                        GetEditor()->DamageRect(devRect);
                        
                        _centerEnds[1]=proj->ConvertFromLocal(localPoint);
                        pts[1]=devPoint;

                        devRect=wxRect(pts[0],pts[1]);
                        GetEditor()->DamageRect(devRect);
                        break;
                }
                case 2: {
                        /* adapt the pavement width */
                        
                        Point3D localFirstPoint=proj->ConvertToLocal(_centerEnds[0]);
                        Point3D localCenter=proj->ConvertToLocal(_geoCenter);
                        Point3D delta=localPoint-localCenter;
                        Point3D lengthwiseVector=(localCenter-localFirstPoint);
                        Point3D constraintVector;
                        constraintVector.setx(-lengthwiseVector.y());
                        constraintVector.sety(lengthwiseVector.x());
                        double len=localFirstPoint.distance3D(localCenter);
                        constraintVector/=len;
                        double deltaL=delta.x()*constraintVector.x()+delta.y()*constraintVector.y();
                        Point3D cstrPoint=localCenter+deltaL*constraintVector;
                        
                        Point3D geoSide;
                        geoSide=proj->ConvertFromLocal(cstrPoint);
                        
                        double alt=_airport->GetElevation()*0.3048;
                        double lat1,lon1;
                        double lat2,lon2;
                        double az1,az2,s;
                        
                        lat1=_geoCenter.lat();
                        lon1=_geoCenter.lon();
                        lat2=geoSide.lat();
                        lon2=geoSide.lon();
                        
                        geo_inverse_wgs_84( alt, lat1, lon1, lat2, lon2, &az1, &az2, &s );
                        
                        _pavementObject->SetWidth(2.0*s/0.3048);
                        break;
                }
        }
        
        // let other modes have it too
        evt.Skip();
}

void PavementInsertionMode::OnLeftUp(wxMouseEvent& evt) {
        if (!_active || evt.AltDown() || evt.ControlDown()) {
                evt.Skip();
                return;
        }
        _active=false;
        MapEditor* editor=GetEditor();
        MapCoordinateSystem& cs=editor->GetCoordinateSystem();
        MapProjection* proj=editor->GetProjection();
        wxPoint devPoint=evt.GetPosition();
        Point3D localPoint=cs.DeviceToLogical(devPoint);
        Point3D geoPoint=proj->ConvertFromLocal(localPoint);
        
        switch (_pointCount) {
                case 0:
                        /* record the first point */
                        _centerEnds[0]=_centerEnds[1]=geoPoint;
                        _pointCount=1;
                        UpdateStatus();
                        break;
                case 1:
                        /* record the second point */
                        _centerEnds[1]=geoPoint;
                        _pointCount=2;
                        UpdateStatus();
                        break;
                case 2: {
                        /* commit and prepare for next pavement segment */
                        TransactionManager* tmgr=GetEditor()->GetTransactionManager();
                        if (_mode==InsertionMode_InsertRunway) {
                                tmgr->Commit(wxT("Inserting Runway"));
                        } else if (_mode==InsertionMode_InsertTaxiway) {
                                tmgr->Commit(wxT("Inserting Taxiway"));
                        }
                        _pointCount=0;
                        return;
                }
        }
        
        if (_pointCount<2)
                return;

        Point3D localFirstPoint=proj->ConvertToLocal(_centerEnds[0]);
        Point3D localCenter=(localFirstPoint+localPoint)/2.0;
        
        _geoCenter=proj->ConvertFromLocal(localCenter);
        
        double alt=_airport->GetElevation()*0.3048;
        double lat1,lon1;
        double lat2,lon2;
        double az1,az2,s;
        
        lat1=_centerEnds[0].lat();
        lon1=_centerEnds[0].lon();
        lat2=_geoCenter.lat();
        lon2=_geoCenter.lon();
        
        geo_inverse_wgs_84( alt, lat1, lon1, lat2, lon2, &az1, &az2, &s );
        
        if (_mode==InsertionMode_InsertRunway) {
                RunwayMapObject::RunwayEndData endData;
                endData.displacedThreshold=0;
                endData.stopwayLength=0;
                endData.vasiAngle=3.0;
                endData.visualApproachPathLighting=RunwayMapObject::VAPL_VASI;
                endData.runwayLighting=RunwayMapObject::RWYL_REIL;
                endData.approachLighting=RunwayMapObject::APPL_ALSF_I;
                _pavementObject=new RunwayMapObject(_airport,
                                                    wxT("xxx"),
                                                    _geoCenter,
                                                    2.0*s/0.3048,
                                                    0.0,
                                                    az1,
                                                    PavementMapObject::Surface_Asphalt,
                                                    0.25,
                                                    endData,
                                                    endData,
                                                    RunwayMapObject::Markings_Visual,
                                                    RunwayMapObject::Shoulder_Asphalt,
                                                    false);
                _airport->GetRunwayLayer()->AttachObject(_pavementObject);
        } else if (_mode==InsertionMode_InsertTaxiway) {
                _pavementObject=new TaxiwayMapObject(_airport,
                                                    _geoCenter,
                                                    2.0*s/0.3048,
                                                    0.0,
                                                    az1,
                                                    PavementMapObject::Surface_Asphalt,
                                                    0.25,
                                                    true,
                                                    true);
                _airport->GetTaxiwayLayer()->AttachObject(_pavementObject);
        }
        
        MapSelectionManager& selmgr=GetEditor()->GetSelectionManager();
        selmgr.DeselectAll();
        selmgr.AddSelectionFor(_pavementObject);
}

