#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include "SelectionMode.h"
#include "ModifyMode.h"
#include <Map/MapEditor.h>
#include <Map/MapObject.h>
#include <Map/MapSelectionHandle.h>
#include <Map/MapSelectionManager.h>
#include <Map/MapCoordinateSystem.h>

#include <algorithm>

IMPLEMENT_CLASS(SelectionMode,Mode)

BEGIN_EVENT_TABLE(SelectionMode,Mode)
        EVT_LEFT_DOWN(SelectionMode::OnLeftButtonDown)
        EVT_MOTION(SelectionMode::OnMouseMotion)
        EVT_LEFT_UP(SelectionMode::OnLeftButtonUp)
        EVT_PAINT(SelectionMode::OnPaint)
END_EVENT_TABLE()

bool SelectionMode::Include_Intersected=true;

SelectionMode::SelectionMode():
        _active(false) {
}

bool SelectionMode::CanExit() {
        return false;
}

void SelectionMode::OnLeftButtonDown(wxMouseEvent& evt) {
        if (evt.AltDown() || evt.ControlDown()) {
                evt.Skip();
                return;
        }

        /* setup stuff for box selection mode */
        _endPoint=_startPoint=evt.GetPosition();
        _mouseMoved=false;
        
        MapSelectionManager& selmgr=GetEditor()->GetSelectionManager();
        MapCoordinateSystem& cs=GetEditor()->GetCoordinateSystem();
        Point3D logicalEndPoint=cs.DeviceToLogical(_endPoint);

        MapObject *hitObject=NULL;
        MapSelection *hitSel=NULL;

        /* if (!hitObject) */
        /*
         * Check whether we have hit a selection handle.
         * We must first check for hitting a handle, as handles may extend
         * beyond the area of the object they belong to and therefore hits
         * to such handles would not be detected when only checking for
         * object hits.
         */
        MapSelectionHandle *hitHandle=selmgr.HitHandle(logicalEndPoint);
        if (hitHandle) {
                hitSel=hitHandle->GetSelection();
                hitObject=hitSel->GetContent();
                // FIXME: properly destroy the handle here
        }
        /* } */
        
        // TODO: actually we should also check for selections in general
        // but we must take care that a selection of an obscured object
        // does still allow to hit the obscuring object
        // The Hit() method of MapSelection should therefore only return true
        // if the actual selection (selection markers+handles) and not the object was hit.
        
        if (!hitObject) {
                /* Check wether we have hit an object */
                TDRectangle hitRect;
                MapObjectList hitObjects;
                hitRect=TDRectangle(logicalEndPoint,logicalEndPoint);
                GetEditor()->GetHitObjects(hitObjects,hitRect);
                
                if (hitObjects.GetCount()!=0) {
                        /* always assume the topmost to be hit */
                        hitObject=hitObjects.GetLast()->GetData();
                        hitSel=selmgr.GetSelectionForObject(hitObject);
                }
        }
        
        if (!hitObject) {
                /* No object hit */
                if (!evt.ShiftDown()) {
                        /* No SHIFT => clear selection */
                        selmgr.DeselectAll();
                }
                /* start box select */
                _active=true;
                return;
        }

        if (hitSel && evt.ShiftDown()) {
                /* we have a selection and SHIFT is pressed
                 * => remove the object's selection and start box select
                 */
                selmgr.RemoveSelection(hitSel);
                _active=true;
                return;
        }
        
        if (!hitSel) {
                /* We have no selection */
                if (!evt.ShiftDown()) {
                        /* SHIFT is not pressed => clear selection */
                        selmgr.DeselectAll();
                }
                selmgr.AddSelectionFor(hitObject);
        }
        
        /* start modify mode */
        _active=false;
        ModifyMode* modmode=new ModifyMode(wxT("Changing Selection"));
        GetEditor()->GetModeManager().InsertOnTop(modmode);
        modmode->ProcessEvent(evt);
}

void SelectionMode::OnMouseMotion(wxMouseEvent& evt) {
        if (!_active || evt.AltDown() || evt.ControlDown()) {
                evt.Skip();
                return;
        }
        wxRect oldRect=GetCurrentDeviceRectangle();
        _endPoint=evt.GetPosition();
        wxRect newRect=GetCurrentDeviceRectangle();
        GetEditor()->DamageRect(oldRect);
        GetEditor()->DamageRect(newRect);
        _mouseMoved=true;
}

void SelectionMode::OnLeftButtonUp(wxMouseEvent& evt) {
        if (!_active || evt.AltDown() || evt.ControlDown()) {
                evt.Skip();
                return;
        }
        _endPoint=evt.GetPosition();
        
        /* Force redraw on the region covered by the selection rectangle */
        GetEditor()->DamageRect(GetCurrentDeviceRectangle());
        
        if (_mouseMoved) {
                TDRectangle hitRect;
                Point3D logicalStartPoint,logicalEndPoint;
                logicalStartPoint=GetEditor()->GetCoordinateSystem().DeviceToLogical(_startPoint);
                logicalEndPoint=GetEditor()->GetCoordinateSystem().DeviceToLogical(_endPoint);
                hitRect=TDRectangle(logicalStartPoint,logicalEndPoint);
                
                MapObjectList list;
                GetEditor()->GetHitObjects(list,hitRect,Include_Intersected);
                
                MapSelectionManager& selmgr=GetEditor()->GetSelectionManager();
                MapObjectList::Node *node;
                for (node=list.GetFirst();node;node=node->GetNext()) {
                        selmgr.AddSelectionFor(node->GetData());
                }
        }
        
        GetEditor()->GetTransactionManager()->Commit(wxT("Changing Selection"));
        
        _active=false;
        _mouseMoved=false;
}

void SelectionMode::OnPaint(wxPaintEvent& evt) {
        if (!_active) {
                evt.Skip();
                return;
        }
        wxClientDC dc(GetEditor()->GetDrawingWindow());
        wxRect rect=GetCurrentDeviceRectangle();
        
        wxPen pen(*wxBLACK,1,wxDOT);
        dc.SetPen(pen);
        dc.SetBrush(*wxTRANSPARENT_BRUSH);
        dc.DrawRectangle(rect);
        dc.SetPen(wxNullPen);
        dc.SetBrush(wxNullBrush);
        
        // Pass on to other modes
        evt.Skip();
}

wxRect SelectionMode::GetCurrentDeviceRectangle() {
        wxCoord top,bottom,right,left;
        
        left=std::min(_startPoint.x,_endPoint.x);
        right=std::max(_startPoint.x,_endPoint.x);
        top=std::min(_startPoint.y,_endPoint.y);
        bottom=std::max(_startPoint.y,_endPoint.y);
        
        return wxRect(left,top,right-left,bottom-top);
}
